"""This submodule contains tools for writing and reading to streams and files.

:synopsis: This submodule contains tools for writing and reading to streams and files.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .handle import *
