"""This submodule provides the necessary objects for parameter estimation using only a cost function
without explicit data and model. Because of this kafe2 Plots are not available, though plots of the
parameter profiles and contours can still be created.

:synopsis: This submodule provides the necessary objects for parameter estimation using only a cost
function without explicit data and model.

.. moduleauthor:: Johannes Gaessler <johannes.gaessler@student.kit.edu>
"""

from .fit import *
